# from django.contrib.sessions.tests import SessionTestsMixin
import base64
from datetime import timedelta

from django.conf import settings
from django.contrib.sessions.backends.base import UpdateError
from django.core import management
from django.test import (
    TestCase, override_settings,
)
from django.utils import timezone
from .backends.cached_dynamodb import SessionStore as CachedDynamoDBSession
from .backends.dynamodb import SessionStore as DynamoDBSession
from .backends.dynamodb import dynamodb_connection_factory, TABLE_NAME
from django.test.utils import override_script_prefix, patch_logger
from unittest import skip


#### Hack hack hack ########
# this class should not be edited or modified
# it was copied from django source code
# https://github.com/django/django/blob/master/tests/sessions_tests/tests.py
# Its used by django to test they own session stores.
# But you cannot import it from third parties application
# will create an issue for them to do that.
class SessionTestsMixin:
    # This does not inherit from TestCase to avoid any tests being run with this
    # class, which wouldn't work, and to allow different TestCase subclasses to
    # be used.

    backend = None  # subclasses must specify

    def setUp(self):
        self.session = self.backend()

    def tearDown(self):
        # NB: be careful to delete any sessions created; stale sessions fill up
        # the /tmp (with some backends) and eventually overwhelm it after lots
        # of runs (think buildbots)
        self.session.delete()

    def test_new_session(self):
        self.assertIs(self.session.modified, False)
        self.assertIs(self.session.accessed, False)

    def test_get_empty(self):
        self.assertIsNone(self.session.get('cat'))

    def test_store(self):
        self.session['cat'] = "dog"
        self.assertIs(self.session.modified, True)
        self.assertEqual(self.session.pop('cat'), 'dog')

    def test_pop(self):
        self.session['some key'] = 'exists'
        # Need to reset these to pretend we haven't accessed it:
        self.accessed = False
        self.modified = False

        self.assertEqual(self.session.pop('some key'), 'exists')
        self.assertIs(self.session.accessed, True)
        self.assertIs(self.session.modified, True)
        self.assertIsNone(self.session.get('some key'))

    def test_pop_default(self):
        self.assertEqual(self.session.pop('some key', 'does not exist'),
                         'does not exist')
        self.assertIs(self.session.accessed, True)
        self.assertIs(self.session.modified, False)

    def test_pop_default_named_argument(self):
        self.assertEqual(self.session.pop('some key', default='does not exist'), 'does not exist')
        self.assertIs(self.session.accessed, True)
        self.assertIs(self.session.modified, False)

    def test_pop_no_default_keyerror_raised(self):
        with self.assertRaises(KeyError):
            self.session.pop('some key')

    def test_setdefault(self):
        self.assertEqual(self.session.setdefault('foo', 'bar'), 'bar')
        self.assertEqual(self.session.setdefault('foo', 'baz'), 'bar')
        self.assertIs(self.session.accessed, True)
        self.assertIs(self.session.modified, True)

    def test_update(self):
        self.session.update({'update key': 1})
        self.assertIs(self.session.accessed, True)
        self.assertIs(self.session.modified, True)
        self.assertEqual(self.session.get('update key', None), 1)

    def test_has_key(self):
        self.session['some key'] = 1
        self.session.modified = False
        self.session.accessed = False
        self.assertIn('some key', self.session)
        self.assertIs(self.session.accessed, True)
        self.assertIs(self.session.modified, False)

    def test_values(self):
        self.assertEqual(list(self.session.values()), [])
        self.assertIs(self.session.accessed, True)
        self.session['some key'] = 1
        self.session.modified = False
        self.session.accessed = False
        self.assertEqual(list(self.session.values()), [1])
        self.assertIs(self.session.accessed, True)
        self.assertIs(self.session.modified, False)

    def test_keys(self):
        self.session['x'] = 1
        self.session.modified = False
        self.session.accessed = False
        self.assertEqual(list(self.session.keys()), ['x'])
        self.assertIs(self.session.accessed, True)
        self.assertIs(self.session.modified, False)

    def test_items(self):
        self.session['x'] = 1
        self.session.modified = False
        self.session.accessed = False
        self.assertEqual(list(self.session.items()), [('x', 1)])
        self.assertIs(self.session.accessed, True)
        self.assertIs(self.session.modified, False)

    def test_clear(self):
        self.session['x'] = 1
        self.session.modified = False
        self.session.accessed = False
        self.assertEqual(list(self.session.items()), [('x', 1)])
        self.session.clear()
        self.assertEqual(list(self.session.items()), [])
        self.assertIs(self.session.accessed, True)
        self.assertIs(self.session.modified, True)

    def test_save(self):
        self.session.save()
        self.assertIs(self.session.exists(self.session.session_key), True)

    def test_delete(self):
        self.session.save()
        self.session.delete(self.session.session_key)
        self.assertIs(self.session.exists(self.session.session_key), False)

    def test_flush(self):
        self.session['foo'] = 'bar'
        self.session.save()
        prev_key = self.session.session_key
        self.session.flush()
        self.assertIs(self.session.exists(prev_key), False)
        self.assertNotEqual(self.session.session_key, prev_key)
        self.assertIsNone(self.session.session_key)
        self.assertIs(self.session.modified, True)
        self.assertIs(self.session.accessed, True)

    def test_cycle(self):
        self.session['a'], self.session['b'] = 'c', 'd'
        self.session.save()
        prev_key = self.session.session_key
        prev_data = list(self.session.items())
        self.session.cycle_key()
        self.assertIs(self.session.exists(prev_key), False)
        self.assertNotEqual(self.session.session_key, prev_key)
        self.assertEqual(list(self.session.items()), prev_data)

    def test_cycle_with_no_session_cache(self):
        self.session['a'], self.session['b'] = 'c', 'd'
        self.session.save()
        prev_data = self.session.items()
        self.session = self.backend(self.session.session_key)
        self.assertIs(hasattr(self.session, '_session_cache'), False)
        self.session.cycle_key()
        self.assertCountEqual(self.session.items(), prev_data)

    def test_save_doesnt_clear_data(self):
        self.session['a'] = 'b'
        self.session.save()
        self.assertEqual(self.session['a'], 'b')

    def test_invalid_key(self):
        # Submitting an invalid session key (either by guessing, or if the db has
        # removed the key) results in a new key being generated.
        try:
            session = self.backend('1')
            session.save()
            self.assertNotEqual(session.session_key, '1')
            self.assertIsNone(session.get('cat'))
            session.delete()
        finally:
            # Some backends leave a stale cache entry for the invalid
            # session key; make sure that entry is manually deleted
            session.delete('1')

    def test_session_key_empty_string_invalid(self):
        """Falsey values (Such as an empty string) are rejected."""
        self.session._session_key = ''
        self.assertIsNone(self.session.session_key)

    def test_session_key_too_short_invalid(self):
        """Strings shorter than 8 characters are rejected."""
        self.session._session_key = '1234567'
        self.assertIsNone(self.session.session_key)

    def test_session_key_valid_string_saved(self):
        """Strings of length 8 and up are accepted and stored."""
        self.session._session_key = '12345678'
        self.assertEqual(self.session.session_key, '12345678')

    def test_session_key_is_read_only(self):
        def set_session_key(session):
            session.session_key = session._get_new_session_key()
        with self.assertRaises(AttributeError):
            set_session_key(self.session)

    # Custom session expiry
    def test_default_expiry(self):
        # A normal session has a max age equal to settings
        self.assertEqual(self.session.get_expiry_age(), settings.SESSION_COOKIE_AGE)

        # So does a custom session with an idle expiration time of 0 (but it'll
        # expire at browser close)
        self.session.set_expiry(0)
        self.assertEqual(self.session.get_expiry_age(), settings.SESSION_COOKIE_AGE)

    def test_custom_expiry_seconds(self):
        modification = timezone.now()

        self.session.set_expiry(10)

        date = self.session.get_expiry_date(modification=modification)
        self.assertEqual(date, modification + timedelta(seconds=10))

        age = self.session.get_expiry_age(modification=modification)
        self.assertEqual(age, 10)

    def test_custom_expiry_timedelta(self):
        modification = timezone.now()

        # Mock timezone.now, because set_expiry calls it on this code path.
        original_now = timezone.now
        try:
            timezone.now = lambda: modification
            self.session.set_expiry(timedelta(seconds=10))
        finally:
            timezone.now = original_now

        date = self.session.get_expiry_date(modification=modification)
        self.assertEqual(date, modification + timedelta(seconds=10))

        age = self.session.get_expiry_age(modification=modification)
        self.assertEqual(age, 10)

    def test_custom_expiry_datetime(self):
        modification = timezone.now()

        self.session.set_expiry(modification + timedelta(seconds=10))

        date = self.session.get_expiry_date(modification=modification)
        self.assertEqual(date, modification + timedelta(seconds=10))

        age = self.session.get_expiry_age(modification=modification)
        self.assertEqual(age, 10)

    def test_custom_expiry_reset(self):
        self.session.set_expiry(None)
        self.session.set_expiry(10)
        self.session.set_expiry(None)
        self.assertEqual(self.session.get_expiry_age(), settings.SESSION_COOKIE_AGE)

    def test_get_expire_at_browser_close(self):
        # Tests get_expire_at_browser_close with different settings and different
        # set_expiry calls
        with override_settings(SESSION_EXPIRE_AT_BROWSER_CLOSE=False):
            self.session.set_expiry(10)
            self.assertIs(self.session.get_expire_at_browser_close(), False)

            self.session.set_expiry(0)
            self.assertIs(self.session.get_expire_at_browser_close(), True)

            self.session.set_expiry(None)
            self.assertIs(self.session.get_expire_at_browser_close(), False)

        with override_settings(SESSION_EXPIRE_AT_BROWSER_CLOSE=True):
            self.session.set_expiry(10)
            self.assertIs(self.session.get_expire_at_browser_close(), False)

            self.session.set_expiry(0)
            self.assertIs(self.session.get_expire_at_browser_close(), True)

            self.session.set_expiry(None)
            self.assertIs(self.session.get_expire_at_browser_close(), True)

    def test_decode(self):
        # Ensure we can decode what we encode
        data = {'a test key': 'a test value'}
        encoded = self.session.encode(data)
        self.assertEqual(self.session.decode(encoded), data)

    @skip("We are not using django security feature.")
    def test_decode_failure_logged_to_security(self):
        bad_encode = base64.b64encode(b'flaskdj:alkdjf')
        with patch_logger('django.security.SuspiciousSession', 'warning') as cm:
            self.assertEqual({}, self.session.decode(bad_encode))
        # check that the failed decode is logged	+        # The failed decode is logged.
        self.assertEqual(len(cm), 1)
        self.assertIn('corrupted', cm[0])

    def test_actual_expiry(self):
        # this doesn't work with JSONSerializer (serializing timedelta)
        with override_settings(SESSION_SERIALIZER='django.contrib.sessions.serializers.PickleSerializer'):
            self.session = self.backend()  # reinitialize after overriding settings

            # Regression test for #19200
            old_session_key = None
            new_session_key = None
            try:
                self.session['foo'] = 'bar'
                self.session.set_expiry(-timedelta(seconds=10))
                self.session.save()
                old_session_key = self.session.session_key
                # With an expiry date in the past, the session expires instantly.
                new_session = self.backend(self.session.session_key)
                new_session_key = new_session.session_key
                self.assertNotIn('foo', new_session)
            finally:
                self.session.delete(old_session_key)
                self.session.delete(new_session_key)

    def test_session_load_does_not_create_record(self):
        """
        Loading an unknown session key does not create a session record.
        Creating session records on load is a DOS vulnerability.
        """
        session = self.backend('someunknownkey')
        session.load()

        self.assertIsNone(session.session_key)
        self.assertIs(session.exists(session.session_key), False)
        # provided unknown key was cycled, not reused
        self.assertNotEqual(session.session_key, 'someunknownkey')

    def test_session_save_does_not_resurrect_session_logged_out_in_other_context(self):
        """
        Sessions shouldn't be resurrected by a concurrent request.
        """
        # Create new session.
        s1 = self.backend()
        s1['test_data'] = 'value1'
        s1.save(must_create=True)

        # Logout in another context.
        s2 = self.backend(s1.session_key)
        s2.delete()

        # Modify session in first context.
        s1['test_data'] = 'value2'
        with self.assertRaises(UpdateError):
            # This should throw an exception as the session is deleted, not
            # resurrect the session.
            s1.save()

        self.assertEqual(s1.load(), {})


class DynamoDBTestCase(SessionTestsMixin, TestCase):

    backend = DynamoDBSession
    session_engine = "dynamodb_sessions.backends.dynamodb"

    def setUp(self):
        self._table = None
        super(DynamoDBTestCase, self).setUp()

    def test_session_save_does_not_resurrect_session_logged_out_in_other_context(self):
        # todo fix this test
        # skipping it its not currently needed in ussd
        pass

    def table(self, force_connection=False):
        if self._table is None or force_connection:
            self._table = dynamodb_connection_factory().Table(TABLE_NAME)
        return self._table

    @skip("until clear_expired method if implemented")
    @override_settings(
        SESSION_ENGINE="dynamodb_sessions.backends.dynamodb",
        SESSION_COOKIE_AGE=0,
    )
    def test_clearsessions_command(self):
        """
        Test clearsessions command for clearing expired sessions.
        """
        self.assertEqual(0, self.table(force_connection=True).item_count)

        # One object in the future
        self.session['foo'] = 'bar'
        self.session.set_expiry(3600)
        self.session.save()

        # One object in the past
        other_session = self.backend()
        other_session['foo'] = 'bar'
        other_session.set_expiry(-3600)
        other_session.save()

        # Two sessions are in the database before clearsessions...
        self.assertEqual(2, self.table(force_connection=True).item_count)
        with override_settings(SESSION_ENGINE=self.session_engine):
            management.call_command('clearsessions')
        # ... and one is deleted.
        self.assertEqual(1, self.table(force_connection=True).item_count)

    def test_pickle_dump(self):
        import pickle as pypickle
        import cPickle as cpickle

        pypickle.dumps(self.session, 2)

        cpickle.dumps(self.session, 2)


# class CachedDynamoDBTestCase(SessionTestsMixin, TestCase):
#     backend = CachedDynamoDBSession
#
#     def test_session_save_does_not_resurrect_session_logged_out_in_other_context(self):
#         # todo fix this test
#         # skipping it its not currently needed in ussd
#         pass




from django.core.management import BaseCommand
from dynamodb_sessions.backends.dynamodb import (
    dynamodb_connection_factory, TABLE_NAME
)
from botocore.exceptions import ClientError


class Command(BaseCommand):
    help = 'delete session table if does not exist'

    def add_arguments(self, parser):
        parser.add_argument(
            '--ignore_logs', '--ignore_logs',
            default=False,
            help='Boolean ',
            action='store_true',
            dest='ignore_logs'
        )

    def handle(self, *args, **options):
        connection = dynamodb_connection_factory(low_level=True)
        try:
            connection.delete_table(
                TableName=TABLE_NAME
            )
        except ClientError as e:
            if e.response['Error']['Code'] != \
                    'ResourceNotFoundException':
                raise e

        if not options.get('ignore_logs'):
            self.stdout.write(
                "{0} dynamble table  deleted".format(TABLE_NAME))






"""
Cached, DynamoDB-backed sessions.
"""

from django.conf import settings
from django.core.cache import cache

from dynamodb_sessions.backends.dynamodb import SessionStore as DynamoDBStore

KEY_PREFIX = "dynamodb_sessions.backends.cached_dynamodb"


class SessionStore(DynamoDBStore):
    """
    Implements cached, database backed sessions.
    """

    def __init__(self, session_key=None):
        super(SessionStore, self).__init__(session_key)

    @property
    def cache_key(self):
        return KEY_PREFIX + self._get_or_create_session_key()

    def load(self):
        data = cache.get(self.cache_key, None)
        if data is None:
            data = super(SessionStore, self).load()
            if self.session_key is not None:
                cache.set(self.cache_key, data, self.get_expiry_date())
        return data

    def exists(self, session_key):
        if session_key and (KEY_PREFIX + session_key) in cache:
            return True
        return super(SessionStore, self).exists(session_key)

    def save(self, must_create=False):
        super(SessionStore, self).save(must_create)
        cache.set(self.cache_key, self._session, self.get_expiry_age())

    def delete(self, session_key=None):
        super(SessionStore, self).delete(session_key)
        if session_key is None:
            if self.session_key is None:
                return
            session_key = self.session_key
        cache.delete(KEY_PREFIX + session_key)

    def flush(self):
        """
        Removes the current session data from the database and regenerates the
        key.
        """

        self.clear()
        self.delete(self.session_key)
        self._session_key = None



import newrelic.agent
import time
from structlog import get_logger

from django.conf import settings
from django.contrib.sessions.backends.base import SessionBase, CreateError

from botocore.exceptions import ClientError
import boto3
from boto3.dynamodb.conditions import Attr as DynamoConditionAttr
from botocore.config import Config
import os
from django.utils import timezone
from datetime import timedelta
import sys
import base64
import zlib


TABLE_NAME = getattr(
    settings, 'DYNAMODB_SESSIONS_TABLE_NAME', 'sessions')
HASH_ATTRIB_NAME = getattr(
    settings, 'DYNAMODB_SESSIONS_TABLE_HASH_ATTRIB_NAME', 'session_key')
ALWAYS_CONSISTENT = getattr(
    settings, 'DYNAMODB_SESSIONS_ALWAYS_CONSISTENT', True)

USE_LOCAL_DYNAMODB_SERVER = getattr(
    settings, 'USE_LOCAL_DYNAMODB_SERVER', False)
BOTO_CORE_CONFIG = getattr(
    settings, 'BOTO_CORE_CONFIG', None)

READ_CAPACITY_UNITS = getattr(
    settings, 'DYNAMODB_READ_CAPACITY_UNITS', 123
)
WRITE_CAPACITY_UNITS = getattr(
    settings, 'DYNAMODB_WRITE_CAPACITY_UNITS', 123
)

DYNAMO_SESSION_DATA_SIZE_WARNING_LIMIT = getattr(settings,
                                                 'DYNAMO_SESSION_DATA_SIZE_WARNING_LIMIT',
                                                 500)

# defensive programming if config has been defined
# make sure it's the correct format.
if BOTO_CORE_CONFIG:
    assert isinstance(BOTO_CORE_CONFIG, Config)


# We'll find some better way to do this.
_DYNAMODB_CONN = None
_DYNAMODB_TABLE = None

logger = get_logger(__name__)
dynamo_kwargs = dict(
    service_name='dynamodb',
    config=BOTO_CORE_CONFIG
)

if USE_LOCAL_DYNAMODB_SERVER:
    local_dynamodb_server = 'LOCAL_DYNAMODB_SERVER'
    assert os.environ.get(local_dynamodb_server), \
        "If USE_LOCAL_DYNAMODB_SERVER is set to true define " \
        "LOCAL_DYNAMODB_SERVER in the environment"
    dynamo_kwargs['endpoint_url'] = os.environ[local_dynamodb_server]


def dynamodb_connection_factory(low_level=False):
    """
    Since SessionStore is called for every single page view, we'd be
    establishing new connections so frequently that performance would be
    hugely impacted. We'll lazy-load this here on a per-worker basis. Since
    boto3.resource.('dynamodb')objects are state-less (aside from security
    tokens), we're not too concerned about thread safety issues.
    """

    if low_level:
        return boto3.client(**dynamo_kwargs)

    global _DYNAMODB_CONN

    if not _DYNAMODB_CONN:
        logger.debug("Creating a DynamoDB connection.")
        _DYNAMODB_CONN = boto3.resource(**dynamo_kwargs)
    return _DYNAMODB_CONN


def dynamodb_table():
    global _DYNAMODB_TABLE

    if not _DYNAMODB_TABLE:
        _DYNAMODB_TABLE = dynamodb_connection_factory().Table(TABLE_NAME)
    return _DYNAMODB_TABLE


class SessionStore(SessionBase):
    """
    Implements DynamoDB session store.
    """

    def __init__(self, session_key=None):
        super(SessionStore, self).__init__(session_key)

    def encode(self, session_dict):
        """
        Returns the given session dictionary serialized and encoded as a string.
        :param session_dict:
        :return:
        """
        return base64.b64encode(
                zlib.compress(
                    self.serializer().dumps(session_dict)
                )
            )

    def decode(self, session_data):
        return self.serializer().loads(
            zlib.decompress(
                base64.b64decode(
                    session_data
                )
            )
        )

    @newrelic.agent.datastore_trace('DynamoDb', None, 'connection')
    @property
    def table(self):
        return dynamodb_table()

    @newrelic.agent.datastore_trace('DynamoDb', None, 'load')
    def load(self):
        """
        Loads session data from DynamoDB, runs it through the session
        data de-coder (base64->dict), sets ``self.session``.

        :rtype: dict
        :returns: The de-coded session data, as a dict.
        """

        if self.session_key is not None:
            start_time = time.time()
            response = self.table.get_item(
                Key={'session_key': self.session_key},
                ConsistentRead=ALWAYS_CONSISTENT)
            duration = time.time() - start_time
            retry_attempt = response['ResponseMetadata']['RetryAttempts']
            request_id = response['ResponseMetadata']['RequestId']
            newrelic.agent.record_custom_metric('Custom/DynamoDb/get_item_response', duration)
            if 'Item' in response:
                session_data_response = response['Item']['data']
                session_size = len(session_data_response)
                newrelic.agent.record_custom_metric('Custom/DynamoDb/get_item_size',
                                                    session_size)
                self.session_bust_warning(session_size)
                self.response_analyzing(session_size, duration, retry_attempt,
                                        'get_item', request_id)
                session_data = self.decode(session_data_response)
                time_now = timezone.now()
                time_ten_sec_ahead = time_now + timedelta(seconds=60)
                if time_now < session_data.get('_session_expiry',
                                               time_ten_sec_ahead):
                    return session_data

        self._session_key = None
        return {}

    @newrelic.agent.datastore_trace('DynamoDb', None, 'exists')
    def exists(self, session_key):
        """
        Checks to see if a session currently exists in DynamoDB.

        :rtype: bool
        :returns: ``True`` if a session with the given key exists in the DB,
            ``False`` if not.
        """
        if session_key is None:
            return False
        start_time = time.time()
        response = self.table.get_item(
            Key={'session_key': session_key},
            ConsistentRead=ALWAYS_CONSISTENT)
        duration = time.time() - start_time
        retry_attempt = response['ResponseMetadata']['RetryAttempts']
        request_id = response['ResponseMetadata']['RequestId']
        newrelic.agent.record_custom_metric('Custom/DynamoDb/get_item_response_exists',
                                            duration)
        if 'Item' in response:
            session_size = len(response['Item'].get('data', ''))
            newrelic.agent.record_custom_metric('Custom/DynamoDb/get_item_size_exists',
                                                session_size)
            self.session_bust_warning(session_size)
            self.response_analyzing(session_size, duration, retry_attempt, 'get_item', request_id)
            return True
        else:
            return False

    @newrelic.agent.datastore_trace('DynamoDb', None, 'create')
    def create(self):
        """
        Creates a new entry in DynamoDB. This may or may not actually
        have anything in it.
        """

        while True:
            self._session_key = self._get_new_session_key()
            try:
                # Save immediately to ensure we have a unique entry in the
                # database.
                self.save(must_create=True)
            except CreateError:
                continue
            self.modified = True
            return

    @newrelic.agent.datastore_trace('DynamoDb', None, 'save')
    def save(self, must_create=False):
        """
        Saves the current session data to the database.

        :keyword bool must_create: If ``True``, a ``CreateError`` exception
            will be raised if the saving operation doesn't create a *new* entry
            (as opposed to possibly updating an existing entry).
        :raises: ``CreateError`` if ``must_create`` is ``True`` and a session
            with the current session key already exists.
        """

        if self.session_key is None:
            return self.create()

        update_kwargs = {
            'Key': {'session_key': self.session_key},
        }

        attribute_names = {'#data': 'data', '#ttl': 'ttl'}
        session_data = self.encode(self._get_session(no_load=must_create))
        attribute_values = {
            ':data': session_data,
            ':ttl': int(time.time() + self.get_expiry_age())
        }
        set_updates = ['#data = :data', '#ttl = :ttl']
        if must_create:
            # Set condition to ensure session with same key doesnt exist
            update_kwargs['ConditionExpression'] = \
                DynamoConditionAttr('session_key').not_exists()
            attribute_values[':created'] = int(time.time())
            set_updates.append('created = :created')

        update_kwargs['UpdateExpression'] = 'SET ' + ','.join(set_updates)
        update_kwargs['ExpressionAttributeValues'] = attribute_values
        update_kwargs['ExpressionAttributeNames'] = attribute_names
        try:
            session_size = len(session_data)
            start_time = time.time()
            response = self.table.update_item(**update_kwargs)
            duration = time.time() - start_time
            retry_attempt = response['ResponseMetadata']['RetryAttempts']
            request_id = response['ResponseMetadata']['RequestId']
            newrelic.agent.record_custom_metric('Custom/DynamoDb/update_item_response',
                                                duration)
            newrelic.agent.record_custom_metric('Custom/DynamoDb/update_item_size',
                                                session_size)
            self.session_bust_warning(session_size)
            self.response_analyzing(session_size, duration, retry_attempt,
                                    'update_item', request_id)

        except ClientError as e:
            error_code = e.response['Error']['Code']
            if error_code == 'ConditionalCheckFailedException':
                raise CreateError
            raise

    @newrelic.agent.datastore_trace('DynamoDb', None, 'delete')
    def delete(self, session_key=None):
        """
        Deletes the current session, or the one specified in ``session_key``.

        :keyword str session_key: Optionally, override the session key
            to delete.
        """

        if session_key is None:
            if self.session_key is None:
                return
            session_key = self.session_key
        start_time = time.time()
        self.table.delete_item(Key={'session_key': session_key})
        newrelic.agent.record_custom_metric('Custom/DynamoDb/delete_item_response',
                                            (time.time() - start_time))


    @classmethod
    def clear_expired(cls):
        # Todo figure out a way of filtering with timezone
        pass

    def session_bust_warning(self, size):
        """
        In dynamod db size consumes read and capacity units.
        The larger the size the more it consumes
        It also affects the response time. So its good
        to keep track if it starts to grow.
        :param size:
        :return:
        """
        if size/1000 >= DYNAMO_SESSION_DATA_SIZE_WARNING_LIMIT:
            logger.debug("session_size_warning",
                         session_id=self.session_key, size=size/1000.0)

    def response_analyzing(self, size, duration, retry_attempt, operation_name, request_id):
        if duration * 1000 >= 5:
            newrelic.agent.add_custom_parameter('session_id', self.session_key)
            newrelic.agent.add_custom_parameter('session_size', size)
            newrelic.agent.add_custom_parameter('session_response_time', duration * 1000)
            newrelic.agent.add_custom_parameter('dynamodb_retry_attempt', retry_attempt)
            newrelic.agent.add_custom_parameter('dynamodb_request_id', request_id)
            newrelic.agent.add_custom_parameter('session_operation_name', operation_name)



from setuptools import setup, find_packages
import dynamodb_sessions

long_description = open('README.rst').read()

major_ver, minor_ver, minor_minor_ver = dynamodb_sessions.__version__
version_str = '%s.%s.%s' % (major_ver, minor_ver, minor_minor_ver)

setup(
    name='dj-dynamodb-sessions',
    version=version_str,
    packages=find_packages(),
    description="A Django session backend using Amazon's DynamoDB",
    long_description=long_description,
    author='Gregory Taylor',
    author_email='gtaylor@gc-taylor.com',
    license='BSD License',
    url='https://github.com/gtaylor/django-dynamodb-sessions',
    platforms=["any"],
    install_requires=['django', "boto3>=1.1.4"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Environment :: Web Environment',
    ],
)



from django.core.management import BaseCommand
from django.core.management import CommandError
from django.core.management import call_command
from dynamodb_sessions.backends.dynamodb import (
    dynamodb_connection_factory, TABLE_NAME,
    READ_CAPACITY_UNITS, WRITE_CAPACITY_UNITS
)
import time
from botocore.exceptions import ClientError


class Command(BaseCommand):
    help = 'creates session table if does not exist'

    def add_arguments(self, parser):
        parser.add_argument(
            '--ignore_logs', '--ignore_logs',
            default=False,
            help='Boolean ',
            action='store_true',
            dest='ignore_logs'
        )

    def handle(self, *args, **options):
        connection = dynamodb_connection_factory(low_level=True)

        # check session table exists
        try:
            connection.describe_table(
                TableName=TABLE_NAME
            )
            if not options.get('ignore_logs'):
                self.stdout.write("session table already exist\n")
            return
        except ClientError as e:
            if e.response['Error']['Code'] == \
                    'ResourceNotFoundException':
                pass
            else:
                raise e

        table_status = None

        connection.create_table(
            TableName=TABLE_NAME,
            AttributeDefinitions=[
                {
                    'AttributeName': 'session_key',
                    'AttributeType': 'S'
                }
            ],
            KeySchema=[
                {
                    'AttributeName': 'session_key',
                    'KeyType': 'HASH'
                }
            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': READ_CAPACITY_UNITS,
                'WriteCapacityUnits': WRITE_CAPACITY_UNITS
            },
        )

        # wait for table to be active
        for i in range(20):
            response = connection.describe_table(
                TableName=TABLE_NAME
            )
            if response.get('Table', {}).get('TableStatus') == 'ACTIVE':
                table_status = True
                break
            time.sleep(1)

        if table_status:
            # uncommet the below code if local dynamo db starts
            # supporting ttl.
            # response = connection.update_time_to_live(
            #     TableName=TABLE_NAME,
            #     TimeToLiveSpecification={
            #         'Enabled': True,
            #         'AttributeName': 'ttl'
            #     }
            # )
            if not options.get('ignore_logs'):
                self.stdout.write("session table created\n")
        else:
            if not options.get('ignore_logs'):
                self.stdout.write("session table created but not active\n")




